{******************************************************************************}
{ Tabela CEST                                                                  }
{                                                                              }
{ Funcionalidades:                                                             }
{ Efetua a leitura do site da lei do convnio onde  criado a CEST e retorna os}
{ cdigos da CEST x NCM e a descrio                                          }
{ https://www.confaz.fazenda.gov.br/legislacao/convenios/2015/cv146_15         }
{                                                                              }
{ Direitos Autorais Reservados (c) 2017 Jter Rabelo Ferreira                  }
{                                                                              }
{  Esta biblioteca  software livre; voc pode redistribu-la e/ou modific-la }
{ sob os termos da Licena Pblica Geral Menor do GNU conforme publicada pela  }
{ Free Software Foundation; tanto a verso 2.1 da Licena, ou (a seu critrio) }
{ qualquer verso posterior.                                                   }
{                                                                              }
{  Esta biblioteca  distribuda na expectativa de que seja til, porm, SEM   }
{ NENHUMA GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU      }
{ ADEQUAO A UMA FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral Menor}
{ do GNU para mais detalhes. (Arquivo LICENA.TXT ou LICENSE.TXT)              }
{                                                                              }
{  Voc deve ter recebido uma cpia da Licena Pblica Geral Menor do GNU junto}
{ com esta biblioteca; se no, escreva para a Free Software Foundation, Inc.,  }
{ no endereo 59 Temple Street, Suite 330, Boston, MA 02111-1307 USA.          }
{ Voc tambm pode obter uma copia da licena em:                              }
{ http://www.opensource.org/licenses/lgpl-license.php                          }
{                                                                              }
{ Jter Rabelo Ferreira  -  jeter.rabelo@jerasoft.com.br                       }
{ Jera Soft Co. - Campestre/MG                                                }
{                                                                              }
{ Desenvolvedor desta unit:                                                    }
{   Lucas da Silva Ferreira -  lucas.ferreira@jerasoft.com.br                  }
{                                                                              }
{******************************************************************************}

unit Jera.LeituraCEST;

interface

uses SysUtils, Generics.Collections, Forms, Dialogs, StrUtils, System.Classes;

type
  TTabelaCEST = class
  private
    FCEST: string;
    FNCM: string;
    FDescricao: string;
  public
    property CEST: string read FCEST write FCEST;
    property NCM: string read FNCM write FNCM;
    property Descricao: string read FDescricao write FDescricao;
  end;

  TTabelaCESTItems = class(TObjectList<TTabelaCEST>) 
  end;

  TLeituraCEST = class
  private
    FItens: TTabelaCESTItems;
    FURL: string;
    procedure DoTratarCapitulos;
    procedure DoTratarEspacoNCM;
    procedure AddCESTCapitulos(const Item: TTabelaCEST);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Ler;
    property URL: string read FURL write FURL;
    property Itens: TTabelaCESTItems read FItens;
  end;

implementation

uses
  System.Types, ACBrUtil, ACBrSocket, Generics.Defaults;

{ TLeituraCEST }

procedure TLeituraCEST.AddCESTCapitulos(const Item: TTabelaCEST);
var
  SCap: TStringDynArray;
  I: Integer;
  NewItem: TTabelaCEST;
  procedure _AddMaisCapitulos;
  var
    IPlus: SmallInt;
    SBegin: string;
    SEnd: string;
    I: Integer;
    SChar: string;
    IBegin, IEnd: SmallInt;
  begin
    IPlus := Pos('+', Item.NCM);
    SBegin := '';
    IBegin := 0;
    for I := IPlus -1 downto 1 do
    begin
      SChar := Trim(Copy(Item.NCM, I, 1));
      if SChar = '' then
        Continue;
      if StrIsNumber(SChar) then
        SBegin := SChar + SBegin
      else
      begin
        IBegin := I;
        Break;
      end;
    end;

    SEnd := '';
    IEnd := Length(Item.NCM);
    for I := IPlus + 1 to Length(Item.NCM) do
    begin
      SChar := Trim(Copy(Item.NCM, I, 1));
      if SChar = '' then
        Continue;
      if StrIsNumber(SChar) then
        SEnd := SEnd + SChar
      else
      begin
        IEnd := I;
        Break;
      end;
    end;
    IEnd := Iend + 1;
    SChar := Item.NCM;
    Delete(SChar, IBegin, IEnd - IBegin);
    Item.NCM := SChar;
    for I := SBegin.ToInteger to SEnd.ToInteger do
    begin
      if Item.NCM <> '' then
        Item.NCM := Item.NCM + '|';
      Item.NCM := Item.NCM + I.ToString;
    end;
  end;
begin
  if Pos('+', Item.NCM) > 0 then
    _AddMaisCapitulos;
  SCap := SplitString(Item.NCM, '|');
  for I := 0 to High(SCap) do
  begin
    NewItem := TTabelaCEST.Create;
    NewItem.CEST := Item.CEST;
    NewItem.NCM := Trim(SCap[I]);
    NewItem.Descricao := Item.Descricao;
    Itens.Add(NewItem);
  end;
end;

constructor TLeituraCEST.Create;
begin
  inherited;
  FItens := TTabelaCESTItems.Create;
  URL := 'https://www.confaz.fazenda.gov.br/legislacao/convenios/2015/cv146_15';
end;

destructor TLeituraCEST.Destroy;
begin
  FreeAndNil(FItens);
  inherited;
end;

procedure TLeituraCEST.DoTratarCapitulos;
var
  I: Integer;
  IConta: Integer;
begin
  IConta := Itens.Count - 1;
  for I := IConta downto 0 do
  begin
    if Copy(LowerCase(Itens[I].NCM), 1, 3) <> 'cap' then
      Continue;
    Itens[I].NCM := LowerCase(TiraAcentos(Itens[I].NCM));
    Itens[I].NCM := Trim(StringReplace(Itens[I].NCM, 'capitulos', '', [rfReplaceAll]));
    Itens[I].NCM := Trim(StringReplace(Itens[I].NCM, 'capitulo', '', [rfReplaceAll]));
    Itens[I].NCM := Trim(StringReplace(Itens[I].NCM, ',', '|', [rfReplaceAll]));
    Itens[I].NCM := Trim(StringReplace(Itens[I].NCM, 'e', '|', [rfReplaceAll]));
    Itens[I].NCM := Trim(StringReplace(Itens[I].NCM, 'a', '+', [rfReplaceAll]));
    if (Pos('|', Itens[I].NCM) = 0) and
       (Pos('+', Itens[I].NCM) = 0)  then
      Continue; // Possui apenas um capitulo
    AddCESTCapitulos(Itens[I]);
    Itens.Delete(I);
  end;
end;

procedure TLeituraCEST.DoTratarEspacoNCM;
var
  I: Integer;
  ONCM: TStringDynArray;
  STmp: string;
  S: Integer;
  NewItem: TTabelaCEST;
begin
  for I := Itens.Count-1 downto 0 do
  begin
    if (Pos(' ', Itens[I].NCM) > 0) and
        (Length(Trim(Itens[I].NCM)) > 0) then
    begin
      STmp := StringReplace(Itens[I].NCM, ' ', '|', [rfReplaceAll]);
      ONcm := SplitString(STmp, '|');
      Itens[I].NCM := ONcm[0];
      for S := 1 to High(ONcm) do
      begin
        NewItem := TTabelaCEST.Create;
        NewItem.CEST := Itens[I].CEST;
        NewItem.NCM := ONcm[S];
        NewItem.Descricao := Itens[I].Descricao;
        Itens.Insert(I +1, NewItem);
      end;
    end;
  end;
end;

procedure TLeituraCEST.Ler;
var
  FLista: TList<string>;
const
  tagInicio = '<table class="plain">';
  tagFim = '</table>';
  trInicio = '<tr>';
  trFim = '</tr>';
  tdInicio = '<td>';
  tdFim = '</td>';
  pInicio = '>';
  pFim = '</p>';

  procedure _AddLista(ATabela: string);
  var
    AQtde: Integer;
    ADeslocamento: Integer;
    AInicio1: Integer;
    AFim1: Integer;
    ALinha: string;
  begin
    // L a tabela completa
    ADeslocamento := 1;
    while (Pos(trInicio, ATabela, ADeslocamento) > 0) do
    begin
      // L a linha
      AInicio1 := Pos(trInicio, ATabela, ADeslocamento);
      AFim1  := Pos(trFim, ATabela, AInicio1);
      AQtde  := (AFim1 - AInicio1) + trFim.Length;
      ADeslocamento := AFim1;
      ALinha := Copy(ATabela, AInicio1, AQtde);
      ALinha := StringReplace(ALinha, '<tr>', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '</tr>', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '</p>', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '<td>', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '<p class="', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '<td colspan="2">', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '">', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '</td>', '|', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '<br> ', '%', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '<br>', '%', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '<span>', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, '</span>', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, 'TabelaSubtitulo', '', [rfReplaceAll]);
      ALinha := StringReplace(ALinha, 'Tabelajustificado', '', [rfReplaceAll]);
      FLista.Add(ALinha);
    end;
  end;

  procedure _CriaOBJs;
    procedure _CarregaLista(ALinha: TStringDynArray);
      procedure _AddLista(ACest: string; ANcm: string; ADescricao: string);
      var
        OTabelaCEST: TTabelaCEST;
      begin
        if UpperCase(ACest) = 'CEST' then
          Exit;
        OTabelaCEST := TTabelaCEST.Create;
        OTabelaCEST.CEST := ACest;
        OTabelaCEST.NCM := ANcm;
        OTabelaCEST.Descricao := ADescricao;
        FItens.Add(OTabelaCEST);
      end;
    var
      SLinhaNCM: TStringDynArray;
      I: Integer;
      SNewLinha: string;
    begin
      SNewLinha := ALinha[2];
      if Pos('', SNewLinha, 2) > 0  then
      begin
        SNewLinha := StringReplace(Copy(SNewLinha, 2, Length(SNewLinha)), '', '%', [rfReplaceAll]);
      end;
      SNewLinha := StringReplace(SNewLinha, '', '', [rfReplaceAll]);
      SLinhaNCM := SplitString(SNewLinha, '%');
      for I := 0 to High(SLinhaNCM) do
      begin
        _AddLista(SplitString(ALinha[1], '')[1],
                  Trim(SLinhaNCM[I]),
                  SplitString(ALinha[3], '')[1]);
      end;
    end;
  var
    I: Integer;
  begin
    for I := 1 to FLista.Count-1 do
    begin
      _CarregaLista(SplitString(FLista[I], '|'));
    end;
  end;
var
  FConteudo: string;
  OHTTP: TACBrHTTP;
  SSite: TStringDynArray;
  I: Integer;
begin
  OHTTP := TACBrHTTP.Create(nil);
  FLista := TList<string>.Create;
  try
    OHTTP.HttpGet(URL);
    FConteudo := OHTTP.RespHTTP.Text;
    FConteudo := StringReplace(FConteudo, #13, '', [rfReplaceAll]);
    FConteudo := StringReplace(FConteudo, #10, '', [rfReplaceAll]);
    FConteudo := StringReplace(FConteudo, tagFim, '', [rfReplaceAll]);
    SSite := SplitString(FConteudo, '');
    for I := 0 to High(SSite) do
      _AddLista(SSite[I]);
    _CriaOBJs;

    // Tratar os capitulos
    DoTratarCapitulos;

    // Tratar as incosistncias do NCM por causa das TAG's do HTML
    DoTratarEspacoNCM;

    // Ordenar o TList
    Itens.Sort(TComparer<TTabelaCEST>.Construct(
          function (const L, R: TTabelaCEST): Integer
          begin
            if L.CEST > R.CEST then
              Result := 1
            else if L.CEST < R.CEST then
              Result := -1
            else
              Result := 0;
          end
        ));
  finally
    FreeAndNil(FLista);
    FreeAndNil(OHTTP);
  end;
end;

end.
